@@ -1,4 +1,4 @@
-# Copyright 2011-2015 MongoDB, Inc.
+# Copyright 2011-present MongoDB, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you
 # may not use this file except in compliance with the License.  You
@@ -16,8 +16,10 @@
 """Functions and classes common to multiple pymongo modules."""
 
 import collections
+import datetime
 import warnings
 
+from bson import SON
 from bson.binary import (STANDARD, PYTHON_LEGACY,
                          JAVA_LEGACY, CSHARP_LEGACY)
 from bson.codec_options import CodecOptions
@@ -27,11 +29,17 @@ from pymongo.auth import MECHANISMS
 from pymongo.errors import ConfigurationError
 from pymongo.monitoring import _validate_event_listeners
 from pymongo.read_concern import ReadConcern
-from pymongo.read_preferences import (read_pref_mode_from_name,
-                                      _ServerMode)
+from pymongo.read_preferences import _MONGOS_MODES, _ServerMode
 from pymongo.ssl_support import validate_cert_reqs
 from pymongo.write_concern import WriteConcern
 
+try:
+    from collections import OrderedDict
+    ORDERED_TYPES = (SON, OrderedDict)
+except ImportError:
+    ORDERED_TYPES = (SON,)
+
+
 # Defaults until we connect to a server and get updated limits.
 MAX_BSON_SIZE = 16 * (1024 ** 2)
 MAX_MESSAGE_SIZE = 2 * MAX_BSON_SIZE
@@ -40,8 +48,9 @@ MAX_WIRE_VERSION = 0
 MAX_WRITE_BATCH_SIZE = 1000
 
 # What this version of PyMongo supports.
-MIN_SUPPORTED_WIRE_VERSION = 0
-MAX_SUPPORTED_WIRE_VERSION = 3
+MIN_SUPPORTED_SERVER_VERSION = "2.6"
+MIN_SUPPORTED_WIRE_VERSION = 2
+MAX_SUPPORTED_WIRE_VERSION = 5
 
 # Frequency to call ismaster on servers, in seconds.
 HEARTBEAT_FREQUENCY = 10
@@ -76,17 +85,19 @@ MAX_IDLE_TIME_MS = None
 # Default value for localThresholdMS.
 LOCAL_THRESHOLD_MS = 15
 
-# mongod/s 2.6 and above return code 59 when a
-# command doesn't exist. mongod versions previous
-# to 2.6 and mongos 2.4.x return no error code
-# when a command does exist. mongos versions previous
-# to 2.4.0 return code 13390 when a command does not
-# exist.
-COMMAND_NOT_FOUND_CODES = (59, 13390, None)
+# Default value for retryWrites.
+RETRY_WRITES = False
+
+# mongod/s 2.6 and above return code 59 when a command doesn't exist.
+COMMAND_NOT_FOUND_CODES = (59,)
 
 # Error codes to ignore if GridFS calls createIndex on a secondary
 UNAUTHORIZED_CODES = (13, 16547, 16548)
 
+# Maximum number of sessions to send in a single endSessions command.
+# From the driver sessions spec.
+_MAX_END_SESSIONS = 10000
+
 
 def partition_node(node):
     """Split a host:port string into (host, int(port)) pair."""
@@ -285,6 +296,14 @@ def validate_timeout_or_zero(option, value):
     return validate_positive_float(option, value) / 1000.0
 
 
+def validate_max_staleness(option, value):
+    """Validates maxStalenessSeconds according to the Max Staleness Spec."""
+    if value == -1 or value == "-1":
+        # Default: No maximum staleness.
+        return -1
+    return validate_positive_integer(option, value)
+
+
 def validate_read_preference(dummy, value):
     """Validate a read preference.
     """
@@ -293,13 +312,16 @@ def validate_read_preference(dummy, value):
     return value
 
 
-def validate_read_preference_mode(dummy, name):
+def validate_read_preference_mode(dummy, value):
     """Validate read preference mode for a MongoReplicaSetClient.
+
+    .. versionchanged:: 3.5
+       Returns the original ``value`` instead of the validated read preference
+       mode.
     """
-    try:
-        return read_pref_mode_from_name(name)
-    except ValueError:
-        raise ValueError("%s is not a valid read preference" % (name,))
+    if value not in _MONGOS_MODES:
+        raise ValueError("%s is not a valid read preference" % (value,))
+    return value
 
 
 def validate_auth_mechanism(option, value):
@@ -381,6 +403,20 @@ def validate_document_class(option, value):
     return value
 
 
+def validate_list(option, value):
+    """Validates that 'value' is a list."""
+    if not isinstance(value, list):
+        raise TypeError("%s must be a list" % (option,))
+    return value
+
+
+def validate_list_or_none(option, value):
+    """Validates that 'value' is a list or None."""
+    if value is None:
+        return value
+    return validate_list(option, value)
+
+
 def validate_is_mapping(option, value):
     """Validate the type of method arguments that expect a document."""
     if not isinstance(value, collections.Mapping):
@@ -398,6 +434,17 @@ def validate_is_document_type(option, value):
                         "collections.MutableMapping" % (option,))
 
 
+def validate_appname_or_none(option, value):
+    """Validate the appname option."""
+    if value is None:
+        return value
+    validate_string(option, value)
+    # We need length in bytes, so encode utf8 first.
+    if len(value.encode('utf-8')) > 128:
+        raise ValueError("%s must be <= 128 bytes" % (option,))
+    return value
+
+
 def validate_ok_for_replace(replacement):
     """Validate a replacement document."""
     validate_is_mapping("replacement", replacement)
@@ -419,6 +466,27 @@ def validate_ok_for_update(update):
         raise ValueError('update only works with $ operators')
 
 
+_UNICODE_DECODE_ERROR_HANDLERS = frozenset(['strict', 'replace', 'ignore'])
+
+
+def validate_unicode_decode_error_handler(dummy, value):
+    """Validate the Unicode decode error handler option of CodecOptions.
+    """
+    if value not in _UNICODE_DECODE_ERROR_HANDLERS:
+        raise ValueError("%s is an invalid Unicode decode error handler. "
+                         "Must be one of "
+                         "%s" % (value, tuple(_UNICODE_DECODE_ERROR_HANDLERS)))
+    return value
+
+
+def validate_tzinfo(dummy, value):
+    """Validate the tzinfo option
+    """
+    if value is not None and not isinstance(value, datetime.tzinfo):
+        raise TypeError("%s must be an instance of datetime.tzinfo" % value)
+    return value
+
+
 # journal is an alias for j,
 # wtimeoutms is an alias for wtimeout,
 URI_VALIDATORS = {
@@ -450,7 +518,10 @@ URI_VALIDATORS = {
     'tz_aware': validate_boolean_or_string,
     'uuidrepresentation': validate_uuid_representation,
     'connect': validate_boolean_or_string,
-    'minpoolsize': validate_non_negative_integer
+    'minpoolsize': validate_non_negative_integer,
+    'appname': validate_appname_or_none,
+    'unicode_decode_error_handler': validate_unicode_decode_error_handler,
+    'retrywrites': validate_boolean_or_string,
 }
 
 TIMEOUT_VALIDATORS = {
@@ -460,12 +531,16 @@ TIMEOUT_VALIDATORS = {
     'serverselectiontimeoutms': validate_timeout_or_zero,
     'heartbeatfrequencyms': validate_timeout_or_none,
     'maxidletimems': validate_timeout_or_none,
+    'maxstalenessseconds': validate_max_staleness,
 }
 
 KW_VALIDATORS = {
     'document_class': validate_document_class,
     'read_preference': validate_read_preference,
-    'event_listeners': _validate_event_listeners
+    'event_listeners': _validate_event_listeners,
+    'tzinfo': validate_tzinfo,
+    'username': validate_string_or_none,
+    'password': validate_string_or_none,
 }
 
 URI_VALIDATORS.update(TIMEOUT_VALIDATORS)
@@ -584,7 +659,8 @@ class BaseObject(object):
 
     @property
     def read_concern(self):
-        """Read only access to the read concern of this instance.
+        """Read only access to the :class:`~pymongo.read_concern.ReadConcern`
+        of this instance.
 
         .. versionadded:: 3.2
         """
