@@ -1,4 +1,4 @@
-# Copyright 2014-2015 MongoDB, Inc.
+# Copyright 2014-present MongoDB, Inc.
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you
 # may not use this file except in compliance with the License.  You
@@ -32,14 +32,16 @@ from pymongo.pool import PoolOptions
 from pymongo.topology_description import (updated_topology_description,
                                           TOPOLOGY_TYPE,
                                           TopologyDescription)
-from pymongo.errors import ServerSelectionTimeoutError
+from pymongo.errors import ServerSelectionTimeoutError, ConfigurationError
 from pymongo.monotonic import time as _time
 from pymongo.server import Server
 from pymongo.server_selectors import (any_server_selector,
-                                      apply_local_threshold,
                                       arbiter_server_selector,
                                       secondary_server_selector,
-                                      writable_server_selector)
+                                      readable_server_selector,
+                                      writable_server_selector,
+                                      Selection)
+from pymongo.client_session import _ServerSessionPool
 
 
 def process_events_queue(queue_ref):
@@ -84,15 +86,17 @@ class Topology(object):
             topology_settings.get_server_descriptions(),
             topology_settings.replica_set_name,
             None,
-            None)
+            None,
+            topology_settings)
 
         self._description = topology_description
         if self._publish_tp:
+            initial_td = TopologyDescription(TOPOLOGY_TYPE.Unknown, {}, None,
+                                             None, None, self._settings)
             self._events.put((
                 self._listeners.publish_topology_description_changed,
-                (TopologyDescription(
-                    TOPOLOGY_TYPE.Unknown, {}, None, None, None),
-                 self._description, self._topology_id)))
+                (initial_td, self._description, self._topology_id)))
+
         for seed in topology_settings.seeds:
             if self._publish_server:
                 self._events.put((self._listeners.publish_server_opened,
@@ -105,6 +109,8 @@ class Topology(object):
         self._condition = self._settings.condition_class(self._lock)
         self._servers = {}
         self._pid = None
+        self._max_cluster_time = None
+        self._session_pool = _ServerSessionPool()
 
         if self._publish_server or self._publish_tp:
             def target():
@@ -128,24 +134,24 @@ class Topology(object):
 
         No effect if called multiple times.
 
-        .. warning:: To avoid a deadlock during Python's getaddrinfo call,
-          will generate a warning if open() is called from a different
-          process than the one that initialized the Topology. To prevent this
-          from happening, MongoClient must be created after any forking OR
-          MongoClient must be started with connect=False.
-        """
-        with self._lock:
-            if self._pid is None:
-                self._pid = os.getpid()
-            else:
-                if os.getpid() != self._pid:
-                    warnings.warn(
-                        "MongoClient opened before fork. Create MongoClient "
-                        "with connect=False, or create client after forking. "
-                        "See PyMongo's documentation for details: http://api."
-                        "mongodb.org/python/current/faq.html#using-pymongo-"
-                        "with-multiprocessing>")
+        .. warning:: Topology is shared among multiple threads and is protected
+          by mutual exclusion. Using Topology from a process other than the one
+          that initialized it will emit a warning and may result in deadlock. To
+          prevent this from happening, MongoClient must be created after any
+          forking.
 
+        """
+        if self._pid is None:
+            self._pid = os.getpid()
+        else:
+            if os.getpid() != self._pid:
+                warnings.warn(
+                    "MongoClient opened before fork. Create MongoClient only "
+                    "after forking. See PyMongo's documentation for details: "
+                    "http://api.mongodb.org/python/current/faq.html#"
+                    "is-pymongo-fork-safe")
+
+        with self._lock:
             self._ensure_opened()
 
     def select_servers(self,
@@ -173,33 +179,41 @@ class Topology(object):
             server_timeout = server_selection_timeout
 
         with self._lock:
-            self._description.check_compatible()
-
-            now = _time()
-            end_time = now + server_timeout
-            server_descriptions = self._apply_selector(selector, address)
-
-            while not server_descriptions:
-                # No suitable servers.
-                if server_timeout == 0 or now > end_time:
-                    raise ServerSelectionTimeoutError(
-                        self._error_message(selector))
-
-                self._ensure_opened()
-                self._request_check_all()
-
-                # Release the lock and wait for the topology description to
-                # change, or for a timeout. We won't miss any changes that
-                # came after our most recent _apply_selector call, since we've
-                # held the lock until now.
-                self._condition.wait(common.MIN_HEARTBEAT_INTERVAL)
-                self._description.check_compatible()
-                now = _time()
-                server_descriptions = self._apply_selector(selector, address)
+            server_descriptions = self._select_servers_loop(
+                selector, server_timeout, address)
 
             return [self.get_server_by_address(sd.address)
                     for sd in server_descriptions]
 
+    def _select_servers_loop(self, selector, timeout, address):
+        """select_servers() guts. Hold the lock when calling this."""
+        now = _time()
+        end_time = now + timeout
+        server_descriptions = self._description.apply_selector(
+            selector, address)
+
+        while not server_descriptions:
+            # No suitable servers.
+            if timeout == 0 or now > end_time:
+                raise ServerSelectionTimeoutError(
+                    self._error_message(selector))
+
+            self._ensure_opened()
+            self._request_check_all()
+
+            # Release the lock and wait for the topology description to
+            # change, or for a timeout. We won't miss any changes that
+            # came after our most recent apply_selector call, since we've
+            # held the lock until now.
+            self._condition.wait(common.MIN_HEARTBEAT_INTERVAL)
+            self._description.check_compatible()
+            now = _time()
+            server_descriptions = self._description.apply_selector(
+                selector, address)
+
+        self._description.check_compatible()
+        return server_descriptions
+
     def select_server(self,
                       selector,
                       server_selection_timeout=None,
@@ -254,6 +268,8 @@ class Topology(object):
                     self._description, server_description)
 
                 self._update_servers()
+                self._receive_cluster_time_no_lock(
+                    server_description.cluster_time)
 
                 if self._publish_tp:
                     self._events.put((
@@ -284,8 +300,7 @@ class Topology(object):
             if topology_type != TOPOLOGY_TYPE.ReplicaSetWithPrimary:
                 return None
 
-            description = writable_server_selector(self._description)[0]
-            return description.address
+            return writable_server_selector(self._new_selection())[0].address
 
     def _get_replica_set_members(self, selector):
         """Return set of replica set member addresses."""
@@ -296,8 +311,7 @@ class Topology(object):
                                      TOPOLOGY_TYPE.ReplicaSetNoPrimary):
                 return set()
 
-            descriptions = selector(self._description)
-            return set([d.address for d in descriptions])
+            return set([sd.address for sd in selector(self._new_selection())])
 
     def get_secondaries(self):
         """Return set of secondary addresses."""
@@ -307,6 +321,28 @@ class Topology(object):
         """Return set of arbiter addresses."""
         return self._get_replica_set_members(arbiter_server_selector)
 
+    def max_cluster_time(self):
+        """Return a document, the highest seen $clusterTime."""
+        return self._max_cluster_time
+
+    def _receive_cluster_time_no_lock(self, cluster_time):
+        # Driver Sessions Spec: "Whenever a driver receives a cluster time from
+        # a server it MUST compare it to the current highest seen cluster time
+        # for the deployment. If the new cluster time is higher than the
+        # highest seen cluster time it MUST become the new highest seen cluster
+        # time. Two cluster times are compared using only the BsonTimestamp
+        # value of the clusterTime embedded field."
+        if cluster_time:
+            # ">" uses bson.timestamp.Timestamp's comparison operator.
+            if (not self._max_cluster_time
+                or cluster_time['clusterTime'] >
+                    self._max_cluster_time['clusterTime']):
+                self._max_cluster_time = cluster_time
+
+    def receive_cluster_time(self, cluster_time):
+        with self._lock:
+            self._receive_cluster_time_no_lock(cluster_time)
+
     def request_check_all(self, wait_time=5):
         """Wake all monitors, wait for at least one to check its server."""
         with self._lock:
@@ -348,6 +384,8 @@ class Topology(object):
             # Mark all servers Unknown.
             self._description = self._description.reset()
             self._update_servers()
+            self._opened = False
+
         # Publish only after releasing the lock.
         if self._publish_tp:
             self._events.put((self._listeners.publish_topology_closed,
@@ -359,6 +397,55 @@ class Topology(object):
     def description(self):
         return self._description
 
+    def pop_all_sessions(self):
+        """Pop all session ids from the pool."""
+        with self._lock:
+            return self._session_pool.pop_all()
+
+    def get_server_session(self):
+        """Start or resume a server session, or raise ConfigurationError."""
+        with self._lock:
+            session_timeout = self._description.logical_session_timeout_minutes
+            if session_timeout is None:
+                # Maybe we need an initial scan? Can raise ServerSelectionError.
+                if self._description.topology_type == TOPOLOGY_TYPE.Single:
+                    if not self._description.has_known_servers:
+                        self._select_servers_loop(
+                            any_server_selector,
+                            self._settings.server_selection_timeout,
+                            None)
+                elif not self._description.readable_servers:
+                    self._select_servers_loop(
+                        readable_server_selector,
+                        self._settings.server_selection_timeout,
+                        None)
+
+            session_timeout = self._description.logical_session_timeout_minutes
+            if session_timeout is None:
+                raise ConfigurationError(
+                    "Sessions are not supported by this MongoDB deployment")
+
+            return self._session_pool.get_server_session(session_timeout)
+
+    def return_server_session(self, server_session, lock):
+        if lock:
+            with self._lock:
+                session_timeout = \
+                    self._description.logical_session_timeout_minutes
+                if session_timeout is not None:
+                    self._session_pool.return_server_session(server_session,
+                                                             session_timeout)
+        else:
+            # Called from a __del__ method, can't use a lock.
+            self._session_pool.return_server_session_no_lock(server_session)
+
+    def _new_selection(self):
+        """A Selection object, initially including all known servers.
+
+        Hold the lock when calling this.
+        """
+        return Selection.from_topology_description(self._description)
+
     def _ensure_opened(self):
         """Start monitors, or restart after a fork.
 
@@ -371,10 +458,10 @@ class Topology(object):
             # Start or restart the events publishing thread.
             if self._publish_tp or self._publish_server:
                 self.__events_executor.open()
-        else:
-            # Restart monitors if we forked since previous call.
-            for server in itervalues(self._servers):
-                server.open()
+
+        # Ensure that the monitors are open.
+        for server in itervalues(self._servers):
+            server.open()
 
     def _reset_server(self, address):
         """Clear our pool for a server and mark it Unknown.
@@ -404,21 +491,6 @@ class Topology(object):
         for server in self._servers.values():
             server.request_check()
 
-    def _apply_selector(self, selector, address):
-        if self._description.topology_type == TOPOLOGY_TYPE.Single:
-            # Ignore the selector.
-            return self._description.known_servers
-        elif address:
-            sd = self._description.server_descriptions().get(address)
-            return [sd] if sd else []
-        elif self._description.topology_type == TOPOLOGY_TYPE.Sharded:
-            return apply_local_threshold(self._settings.local_threshold_ms,
-                                         self._description.known_servers)
-        else:
-            sds = selector(self._description)
-            return apply_local_threshold(
-                self._settings.local_threshold_ms, sds)
-
     def _update_servers(self):
         """Sync our Servers from TopologyDescription.server_descriptions.
 
@@ -467,8 +539,8 @@ class Topology(object):
             socket_timeout=options.connect_timeout,
             ssl_context=options.ssl_context,
             ssl_match_hostname=options.ssl_match_hostname,
-            socket_keepalive=True,
-            event_listeners=options.event_listeners)
+            event_listeners=options.event_listeners,
+            appname=options.appname)
 
         return self._settings.pool_class(address, monitor_pool_options,
                                          handshake=False)
